=pod

=head1 NAME

  Bio::EnsEMBL::Hive::Examples::Factories::PipeConfig::CompressFiles5_conf

=head1 SYNOPSIS

    init_pipeline.pl Bio::EnsEMBL::Hive::Examples::Factories::PipeConfig::CompressFiles4_conf -password <your_password>

    seed_pipeline.pl -url <url> -logic_name find_files -input_id "{ 'directory' => 'dumps', 'only_files' => '*.sql' }"

    seed_pipeline.pl -url <url> -logic_name find_files -input_id "{ 'directory' => 'dumps', 'only_files' => '*.gz', 'gzip_flags' => '-d' }"

    seed_pipeline.pl -url <url> -logic_name find_files -input_id "{ 'directory' => '$HOME/ncbi_taxonomy' }"

=head1 DESCRIPTION

    This is an example pipeline put together from five basic building blocks:

    Analysis_1: JobFactory.pm is used to turn the list of files in a given directory into jobs

        these jobs are sent down the branch #2 into the second analysis

    Analysis_2: JobFactory.pm is used to run a wc command to determine the size of a file
                (and format the output with sed), then capture the command's object, putting
                the file size into a parameter for later use.

    Analysis_3: SystemCmd.pm is used to run these compression/decompression jobs in parallel.

    Analysis_4: JobFactory.pm is used to run a wc command to determine the size of a file
                (and format the output with sed), then capture the command's object, putting
                the file size into a parameter for later use.

    Analysis_5: SystemCmd.pm is used to run the notify-send command, displaying a message on the screen.

=head1 LICENSE

    Copyright [1999-2015] Wellcome Trust Sanger Institute and the EMBL-European Bioinformatics Institute
    Copyright [2016-2017] EMBL-European Bioinformatics Institute

    Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software distributed under the License
    is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and limitations under the License.

=head1 CONTACT

    Please subscribe to the Hive mailing list:  http://listserver.ebi.ac.uk/mailman/listinfo/ehive-users  to discuss Hive-related questions or to be notified of our updates

=cut

package Bio::EnsEMBL::Hive::Examples::Factories::PipeConfig::CompressFiles5_conf;

use strict;
use warnings;

use base ('Bio::EnsEMBL::Hive::PipeConfig::HiveGeneric_conf');  # All Hive databases configuration files should inherit from HiveGeneric, directly or indirectly
use Bio::EnsEMBL::Hive::PipeConfig::HiveGeneric_conf; # Need to use as well as use base() to turn on INPUT_PLUS()

=head2 pipeline_analyses

    Description : Implements pipeline_analyses() interface method of Bio::EnsEMBL::Hive::PipeConfig::HiveGeneric_conf that defines the structure of the pipeline: analyses, jobs, rules, etc.
                  Here it defines five analyses:

                    * 'find_files'          generates a list of files whose names match the pattern #only_files#
                                            Each job of this analysis will dataflow (create jobs) via branch #2 into 'compress_a_file' analysis.

                    * 'pre_compress_size'   determines the size of a file, and captures the size in a parameter to pass through the pipeline

                    * 'compress_a_file'     actually performs the (un)zipping of the files in parallel

                    * 'post_compress_size'   determines the size of a file, and captures the size in a parameter to pass through the pipeline

                    * 'notify'              displays a message on the screen using the notify-send command

=cut

sub pipeline_analyses {
    my ($self) = @_;
    return [
        {   -logic_name => 'find_files',
            -module     => 'Bio::EnsEMBL::Hive::RunnableDB::JobFactory',
            -parameters => {
                'inputcmd'     => 'find #directory# -type f -name "#only_files#"',
                'column_names' => [ 'filename' ],
                'directory'     => 'pdfs',  # default directory if one is not otherwise specified
                'only_files'    => '*',     # any wildcard understood by shell
            },
            -flow_into => {
                # Create a fan of jobs, using INPUT_PLUS() to propigate all of
                # this analysis' parameters down branch 2
                '2->A' => { 'pre_compress_size' => INPUT_PLUS() },
                'A->1' => [ 'notify' ],
            },
        },

        {   -logic_name    => 'pre_compress_size',
            -module        => 'Bio::EnsEMBL::Hive::RunnableDB::JobFactory', # to capture output use JobFactory
            -parameters    => {
                'inputcmd'      => "wc -c #filename# | sed -e 's/^ *//' ",
                'delimiter'     => ' ',
                'column_names'  => [ 'orig_size', 'orig_filename' ],
            },
            -flow_into => {
                2 => [ '?accu_name=orig_size&accu_address={orig_filename}' ],
                1 => [ 'compress_a_file' ],
            },
        },

        {   -logic_name    => 'compress_a_file',
            -module        => 'Bio::EnsEMBL::Hive::RunnableDB::SystemCmd',
            -parameters    => {
                'cmd'           => 'gzip #gzip_flags# #filename#',

                # defaults for the parameters used in this analysis:
                'gzip_flags'    => '',      # can be set to '-d' for decompression
            },
            -analysis_capacity => 4,            # limit the number of workers that will be performing jobs in parallel
        },

        {   -logic_name    => 'post_compress_size',
            -module        => 'Bio::EnsEMBL::Hive::RunnableDB::JobFactory', # to capture output use JobFactory
            -parameters    => {
                'inputcmd'      => "wc -c #filename#.gz | sed -e 's/^ *//' ",
                'delimiter'     => ' ',
                'column_names'  => [ 'comp_size', 'comp_filename' ],
            },
            -flow_into => {
                2 => [ '?accu_name=comp_size&accu_address={comp_filename}' ],
            },
        },
        {   -logic_name    => 'notify',
            -module        => 'Bio::EnsEMBL::Hive::RunnableDB::SystemCmd',
            -meadow_type   => 'LOCAL',
            -parameters    => {
                'text' => "compressed file sizes between #min_comp_size# and #max_comp_size#",
                'cmd' => 'notify-send "#text#"',
                'min_comp_size' => '#expr(min values %{#comp_size#})expr#',
                'max_comp_size' => '#expr(max values %{#comp_size#})expr#',
            },
        }
    ];
}

1;
