=pod

=head1 NAME

    CompressAndReport

=head1 SYNOPSIS

    standaloneJob.pl CompressAndReport --filename 'myfile.txt.gz' --gzip_flags '-d'

=head1 DESCRIPTION

    This runnable [de]compresses a particular file, and makes note of the size
    of that file before and after [de]compression

    This operation could also be done by chaining together steps using the default
    SystemCmd Runable; here they are all combined into a single Runnable
    as a demonstration.

=head1 LICENSE

    Copyright [1999-2015] Wellcome Trust Sanger Institute and the EMBL-European Bioinformatics Institute
    Copyright [2016-2017] EMBL-European Bioinformatics Institute

    Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

         http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software distributed under the License
    is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and limitations under the License.

=head1 CONTACT

    Please subscribe to the Hive mailing list:  http://listserver.ebi.ac.uk/mailman/listinfo/ehive-users  to discuss Hive-related questions or to be notified of our updates

=cut


package CompressAndReport;

use strict;
use warnings;

use base ('Bio::EnsEMBL::Hive::Process');

sub param_defaults {
    return {'gzip_flags' => ""};
}

sub pre_cleanup {
    # This is where the runnable would take care of getting things
    # (typically files or database connections) in order before starting to run.
    # This is often useful if the jobs sometimes need to be retried before
    # completing successfully
}

sub fetch_input {
    my $self = shift;
    # read_input is typically used to validate input, test db connections are
    # working, etc. Here, we'll check that we were passed a parameter called
    # 'filename' and that it points to a file

    # use param_required for this, since we want the job to fail if there's no filename.
    my $filename = $self->param_required('filename');

    unless (-e $filename) {
        die 'File $filename does not exist'; # Will cause job to fail and leave a message in log_message
    }

}

sub run {
    my $self = shift;

    #here, we have to re-fetch the filename, since $filename went out of scope,
    #this is standard Perl behavior.
    my $filename = $self->param_required('filename');

    # gzip_flags are optional - you can still run this job with no flags -
    # so we want the job to continue even if this param is uninitialized.
    # Therefore, use param to get the parameter.
    my $gzip_flags = $self->param('gzip_flags');

    my $presize = -s $filename;
    my $compress_cmd = "gzip $gzip_flags $filename";
    my ($return_value, $stderr, $flat_cmd) = $self->run_system_command($compress_cmd);
    if ($return_value) {
        die "there was an error running gzip as $flat_cmd: $stderr";
    }

    my $post_operation_filename = _create_post_operation_filename($filename, $gzip_flags);

    my $postsize = -s $post_operation_filename;

    #package up presize and postsize so that write_output will be able to see them

    $self->param('presize', $presize);
    $self->param('postsize', $postsize);
}

sub write_output {
    my $self = shift;

    #here, we flow out three parameters on branch 2: filename, size, and comp_size
    $self->dataflow_output_id({'filename' => $self->param('filename'),
                               'orig_size' => $self->param('presize'),
                               'comp_size' => $self->param('postsize')},
                               2);

    #note that there's also an event flown out on branch 1 by default, because
    #we haven't overridden it.
}

sub post_cleanup {

}

# private method
# if the initial filename had a .gz
sub _create_post_operation_filename {
    my ($pre_operation_filename, $gzip_flags) = @_;

    if ($gzip_flags =~ /-d/) {
        return substr($pre_operation_filename, 0, -3);
    } else {
        return $pre_operation_filename . ".gz";
    }

}

1;
